"""Class implementation for append line point interface.
"""

from typing import Union

import apysc as ap
from apysc._display.points_2d_interface import Points2DInterface


class AppendLinePointInterface(Points2DInterface):

    _points_var_name: str

    def append_line_point(
            self, x: Union[int, ap.Int], y: Union[int, ap.Int]) -> None:
        """
        Append line point at the end.

        Parameters
        ----------
        x : int or Int
            X-coordinate.
        y : int or Int
            Y-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.append_line_point, locals_=locals(),
                module_name=__name__, class_=AppendLinePointInterface):
            from apysc._type import value_util
            if not hasattr(self, '_points_var_name'):
                raise AttributeError(
                    '_points_var_name attribute is not set. Please add '
                    'implementation to set that value when constructor '
                    'or else.')
            self.points.append(value=ap.Point2D(x=x, y=y))
            expression: str
            x_name: str = value_util.get_value_str_for_expression(value=x)
            y_name: str = value_util.get_value_str_for_expression(value=y)
            expression = (
                f'{self._points_var_name}.push([{x_name}, {y_name}]);'
                f'\n{self.variable_name}.plot({self._points_var_name});'
            )
            ap.append_js_expression(expression=expression)



"""Class implementation for the line alpha value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

import apysc as ap
from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationLineAlpha(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a line alpha.
    """

    _line_alpha: ap.Number

    def __init__(
            self,
            target: _T,
            alpha: Union[float, ap.Number],
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a line alpha.

        Parameters
        ----------
        target : VariableNameInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        alpha : float or Number
            The final line alpha of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationLineAlpha):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(
                    type_name=var_names.ANIMATION_LINE_ALPHA)
            self._line_alpha = to_apysc_val_from_builtin.\
                get_copied_number_from_builtin_val(float_or_num=alpha)
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(AnimationLineAlpha, self).__init__(
                variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        line_alpha_str: str = value_util.get_value_str_for_expression(
            value=self._line_alpha)
        return f'\n  .attr({{"stroke-opacity": {line_alpha_str}}});'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        from apysc._display.line_alpha_interface import LineAlphaInterface
        expression: str = ''
        if isinstance(self._target, LineAlphaInterface):
            self._target._initialize_line_alpha_if_not_initialized()
            expression = (
                f'{self._target._line_alpha.variable_name} = '
                f'{self._line_alpha.variable_name};'
            )
        return expression



"""Elif (else if) branch instruction implementations.
"""

from typing import Any
from typing import Dict
from typing import Optional

import apysc as ap
from apysc._branch.if_base import IfBase


class Elif(IfBase):
    """
    A class to append elif branch instruction expression.

    References
    ----------
    - Elif document
        - https://simon-ritchie.github.io/apysc/elif.html
    - Each branch instruction class scope variables reverting setting
        - https://bit.ly/3rkAuaT
    """

    def __init__(
            self,
            condition: Optional[ap.Boolean],
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append elif branch instruction expression.

        Parameters
        ----------
        condition : Boolean or None
            Boolean value to be used for judgment.
        locals_ : dict or None, default None
            Current scope's local variables. Set locals() value to
            this argument. If specified, all local scope
            VariableNameInterface variables (like Int, Sprite) will be
            reverted ad the end of If scope. This setting is useful
            when you don't want to update each variable by the
            implementation of the If scope.
        globals_ : dict or None, default None
            Current scope's global variables. Set golobals() value
            to this argument. This works the same way as the locals_
            argument.

        References
        ----------
        - Elif document
            - https://simon-ritchie.github.io/apysc/elif.html
        - Each branch instruction class scope variables reverting setting
            - https://bit.ly/3rkAuaT
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Elif):
            super(Elif, self).__init__(
                condition=condition, locals_=locals_, globals_=globals_)

    def _append_enter_expression(self) -> None:
        """
        Append else if branch instruction start expression.

        Raises
        ------
        ValueError
            If the last scope is not If or Elif.
        """
        import apysc as ap
        if not self._last_scope_is_if_or_elif():
            raise ValueError(
                'Elif interface can only use right after If or Elif '
                'interfaces.'
                '\n\nMaybe you are using Int or String, or anything else'
                ' comparison expression at Elif constructor (e.g., '
                '`with Elif(any_value == 10, ...):`).'
                '\nCurrently that specifying expression directly is not '
                'supported so please'
                ' define condition seperately as follows:'
                '\ncondition: Boolean = any_value == 10'
                '\n...'
                '\nwith Elif(condition, ....):')
        if self._condition is None:
            raise ValueError(
                'Elif expression\'s condition argument can\'t be set None.')
        expression: str = (
            f'else if ({self._condition.variable_name}) {{'
        )
        ap.append_js_expression(expression=expression)

    def _set_last_scope(self) -> None:
        """
        Set expression last scope value.
        """
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        last_scope.set_last_scope(value=LastScope.ELIF)



"""Class implementation for the attribute linking interface.

This interface is used for updating an old property value to
achieve consistency in the handler functions.
"""

from typing import Dict
from typing import List
from typing import Union

import apysc as ap

_AttrName = str
_Attr = Union[ap.Int, ap.Number, ap.String, ap.Boolean]


class AttrLinkingInterface:

    _attr_linking_stack: Dict[_AttrName, List[_Attr]]

    def _initialize_attr_linking_stack(self, attr_name: str) -> None:
        """
        Initialize the _attr_linking_stack attribute if it hasn't been
        initialized yet.

        Parameters
        ----------
        attr_name : str
            Target attribute name.
        """
        if not hasattr(self, '_attr_linking_stack'):
            self._attr_linking_stack = {}
        if attr_name in self._attr_linking_stack:
            return
        self._attr_linking_stack[attr_name] = []

    def _append_attr_to_linking_stack(
            self, attr: _Attr, attr_name: str) -> None:
        """
        Append an attribute to the linking attribute stack.

        Parameters
        ----------
        attr : Int or Number or String or Boolean
            Target attribute to be appended.
        attr_name : str
            Target attribute name.
        """
        self._initialize_attr_linking_stack(attr_name=attr_name)
        if self._is_target_attr_already_linked(
                attr=attr, attr_name=attr_name):
            return
        self._attr_linking_stack[attr_name].append(attr)

    def _is_target_attr_already_linked(
            self, attr: _Attr, attr_name: str) -> bool:
        """
        Get a boolean value whether a specified attribute has already
        been appended to the linking attribute stack.

        Parameters
        ----------
        attr : Int or Number or String or Boolean
            Target attribute to be appended.
        attr_name : str
            Target attribute name.

        Returns
        -------
        result : bool
            If a specified attribute has already been appended to
            the linking stack, this value will be True.
        """
        self._initialize_attr_linking_stack(attr_name=attr_name)
        for in_stack_value in self._attr_linking_stack[attr_name]:
            if in_stack_value.variable_name == attr.variable_name:
                return True
        return False

    def _append_applying_new_attr_val_exp(
            self, new_attr: _Attr, attr_name: str) -> None:
        """
        Append the expression of applying new attribute value to each
        stacked value.

        Parameters
        ----------
        new_attr : Int or Number or String or Boolean
            New attribute value.

        attr_name : str
            Target attribute name.
        """
        self._initialize_attr_linking_stack(attr_name=attr_name)
        if not self._attr_linking_stack:
            return
        new_attr_name: str = new_attr.variable_name
        expression: str = ''
        for stacked_value in self._attr_linking_stack[attr_name]:
            if stacked_value.variable_name == new_attr_name:
                continue
            if expression != '':
                expression += '\n'
            expression += (
                f'{stacked_value.variable_name} = {new_attr_name};'
            )
        ap.append_js_expression(expression=expression)



"""Class implementation for the flip_x interface.
"""

from typing import Dict

import apysc as ap
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class FlipXInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _flip_x: ap.Boolean

    def _initialize_flip_x_if_not_initialized(self) -> None:
        """
        Initialize the _flip_x attribute if it hasn't been
        initialized yet.
        """
        if hasattr(self, '_flip_x'):
            return
        self._flip_x = ap.Boolean(False)

        self._append_flip_x_attr_linking_setting()

    def _append_flip_x_attr_linking_setting(self) -> None:
        """
        Append a flip-x attribute linking setting.
        """
        with ap.DebugInfo(
                callable_=self._append_flip_x_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._flip_x, attr_name='flip_x')
            self._append_attr_to_linking_stack(
                attr=self._flip_x, attr_name='flip_x')

    @property
    def flip_x(self) -> ap.Boolean:
        """
        Get a boolean value whether the x-axis is flipped or not.

        Returns
        -------
        flip_x : Boolean
            A boolean value whether the x-axis is flipped or not.
        """
        with ap.DebugInfo(
                callable_='flip_x', locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            from apysc._type import value_util
            self._initialize_flip_x_if_not_initialized()
            return value_util.get_copy(value=self._flip_x)

    @flip_x.setter
    def flip_x(self, value: ap.Boolean) -> None:
        """
        Update a x-axis flipping value.

        Parameters
        ----------
        value : Boolean
            Flipping value. If True, a x-axis will be flipped,
            otherwise it will be reset.
        """
        with ap.DebugInfo(
                callable_='flip_x', locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            self._initialize_flip_x_if_not_initialized()
            before_value: ap.Boolean = self._flip_x
            self._flip_x = value
            self._append_flip_x_update_expression(before_value=before_value)

            self._append_flip_x_attr_linking_setting()

    def _append_flip_x_update_expression(
            self, before_value: ap.Boolean) -> None:
        """
        Append a x-axis flipping value updating expression.

        Parameters
        ----------
        before_value : Boolean
            Before updating flipping value.
        """
        with ap.DebugInfo(
                callable_=self._append_flip_x_update_expression,
                locals_=locals(),
                module_name=__name__, class_=FlipXInterface):
            from apysc._display import flip_interface_helper
            self._initialize_flip_x_if_not_initialized()
            expression: str = flip_interface_helper.\
                make_flip_update_expression(
                    before_value=before_value, after_value=self._flip_x,
                    axis=flip_interface_helper.Axis.X,
                    interface_variable_name=self.variable_name)
            ap.append_js_expression(expression=expression)

    _flip_x_snapshots: Dict[str, bool]

    def _make_snapshot(self, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not hasattr(self, '_flip_x_snapshots'):
            self._flip_x_snapshots = {}
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_flip_x_if_not_initialized()
        self._flip_x_snapshots[snapshot_name] = self._flip_x._value

    def _revert(self, snapshot_name: str) -> None:
        """
        Revert a value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._flip_x._value = self._flip_x_snapshots[snapshot_name]



"""Base class implementation for the animation.
"""

from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

import apysc as ap
from apysc._animation.easing import Easing
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)
_O = TypeVar('_O')
_AnimationEvent = Any
_Handler = Callable[[_AnimationEvent, _O], None]


class AnimationBase(
        VariableNameInterface, CustomEventInterface, Generic[_T], ABC):

    _target: _T
    _duration: ap.Int
    _delay: ap.Int
    _easing: Easing
    _started: ap.Boolean

    def __init__(self, variable_name: str) -> None:
        """
        Base class for each animation setting.

        Parameters
        ----------
        variable_name : str
            Variable name.
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationBase):
            self.variable_name = variable_name
            self._started = ap.Boolean(False)

    @abstractmethod
    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.
        """

    def _set_basic_animation_settings(
            self,
            target: _T,
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        Set the basic animation settings.

        Parameters
        ----------
        target : VariableNameInterface
            A target instance of the animation target
            (e.g., `DisplayObject` instance).
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting. If None, Linear calculation is used.
        """
        with ap.DebugInfo(
                callable_=self._set_basic_animation_settings,
                locals_=locals(),
                module_name=__name__, class_=AnimationBase):
            from apysc._converter import to_apysc_val_from_builtin
            self._target = target
            self._duration = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=duration)
            self._delay = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=delay)
            self._easing = easing

    def _get_animation_basic_expression(self) -> str:
        """
        Get an animation basic expression string.

        Returns
        -------
        expression : str
            An animation basic expression string.
        """
        expression: str = (
            f'{self._target.variable_name}'
            '\n  .animate({'
            f'\n    duration: {self._duration.variable_name},'
            f'\n    delay: {self._delay.variable_name}}})'
            f'\n  .ease({self._easing.value})'
        )
        expression += self._get_animation_complete_handler_expression()
        return expression

    def start(self) -> 'AnimationBase':
        """
        Start an animation with current settings.

        Returns
        -------
        self : AnimatonBase
            This instance.

        References
        ----------
        - AnimationBase class start interface
            - https://simon-ritchie.github.io/apysc/animation_base_start.html
        """
        with ap.DebugInfo(
                callable_=self.start, locals_=locals(),
                module_name=__name__, class_=AnimationBase):
            expression: str = self._get_animation_basic_expression()
            animation_expresssion: str = self._get_animation_func_expression()
            expression += animation_expresssion
            ap.append_js_expression(expression=expression)
            self._started.value = True
        return self

    def _get_animation_complete_handler_expression(self) -> str:
        """
        Get a expression of the animation complete handlers setting.

        Returns
        -------
        expression : str
            Target expression string.
            e.g., '\n  .after(handler_name)'

        Notes
        -----
        If multiple handlers are registered, then multiple lines
        expression will be returned.
        """
        from apysc._event.custom_event_type import CustomEventType
        event_type: str = CustomEventType.ANIMATION_COMPLETE.value
        self._initialize_custom_event_handlers_if_not_initialized(
            custom_event_type_str=event_type)
        if event_type not in self._custom_event_handlers:
            return ''
        expression: str = ''
        for handler_name in self._custom_event_handlers[event_type].keys():
            expression += f'\n  .after({handler_name})'
        return expression

    def animation_complete(
            self, handler: _Handler[_O],
            options: Optional[_O] = None) -> 'AnimationBase':
        """
        Add a animation complete event listener setting.

        Notes
        -----
        This interface can only be used before an animation start.

        Parameters
        ----------
        handler : _Handler
            A callable will be called when an animation is complete.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.

        Returns
        -------
        self : AnimatonBase
            This instance.

        Raises
        ------
        Exception
            If this interface is called after an animation start.

        References
        ----------
        - AnimationBase class animation_complete interface document
            - https://simon-ritchie.github.io/apysc/animation_complete.html
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        with ap.DebugInfo(
                callable_=self.animation_complete, locals_=locals(),
                module_name=__name__, class_=AnimationBase):
            from apysc._event.custom_event_type import CustomEventType
            from apysc._validation.handler_options_validation import \
                validate_options_type
            self._validate_animation_not_started()
            validate_options_type(options=options)
            e: ap.AnimationEvent[_T] = ap.AnimationEvent(this=self)
            in_handler_head_expression: str = \
                self._get_complete_event_in_handler_head_expression()
            self.bind_custom_event(
                custom_event_type=CustomEventType.ANIMATION_COMPLETE,
                handler=handler,
                e=e,
                options=options,
                in_handler_head_expression=in_handler_head_expression)
            return self

    @abstractmethod
    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.
        """

    def _validate_animation_not_started(self) -> None:
        """
        Validate whether an animation hasn't already been started.

        Raises
        ------
        Exception
            If an animation has already been started.
        """
        if not self._started._value:
            return
        raise Exception(
            'This interface can not be called after the animation '
            'is started.')

    @property
    def target(self) -> _T:
        """
        Get an animation target instance.

        Returns
        -------
        target : VariableNameInterface
            An animation target instance.

        References
        ----------
        - AnimationBase class target property interface document
            - https://simon-ritchie.github.io/apysc/animation_base_target.html
        """
        return self._target



"""Class implementation for the continue.
"""


class Continue:
    """
    The loop continue expression class.

    Notes
    -----
    This class can be instantiated in the with loop statement,
    for example, after the `with ap.For(...):` statement.
    """

    def __init__(self) -> None:
        """
        The loop continue expression class.

        Notes
        -----
        This class can be instantiated in the with loop statement,
        for example, after the `with ap.For(...):` statement.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Continue):
            from apysc._loop import loop_count
            current_loop_count: int = loop_count.get_current_loop_count()
            if current_loop_count == 0:
                err_msg: str = (
                    'The `Continue` class can be instantiated in the with '
                    'loop statement, for example, after the '
                    '`with ap.For(...):` statement.'
                )
                raise Exception(err_msg)
            ap.append_js_expression(expression='continue;')



"""Round dot setting class implementation for line.
"""

from typing import Union

import apysc as ap


class LineRoundDotSetting(ap.Dictionary[str, ap.Int]):
    """
    Round dot setting class for a line.

    References
    ----------
    - Graphics line_style interface document
        - https://bit.ly/3zauILT
    """

    def __init__(
            self, round_size: Union[int, ap.Int],
            space_size: Union[int, ap.Int]) -> None:
        """
        Round dot setting class for line.

        Parameters
        ----------
        round_size : int or Int
            Dot round size.
        space_size : int or Int
            Blank space size between dots.

        References
        ----------
        - Graphics line_style interface document
            - https://bit.ly/3zauILT
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=LineRoundDotSetting):
            from apysc._converter.to_apysc_val_from_builtin import \
                get_copied_int_from_builtin_val
            from apysc._validation import number_validation
            number_validation.validate_nums_are_int_and_gt_zero(
                nums=[round_size, space_size])
            round_size_: ap.Int = get_copied_int_from_builtin_val(
                integer=round_size)
            space_size_: ap.Int = get_copied_int_from_builtin_val(
                integer=space_size)
            super(LineRoundDotSetting, self).__init__({
                'round_size': round_size_,
                'space_size': space_size_,
            })

    @property
    def round_size(self) -> ap.Int:
        """
        Get a round size setting.

        Returns
        -------
        round_size : Int
            Round size setting.
        """
        with ap.DebugInfo(
                callable_='round_size', locals_=locals(),
                module_name=__name__, class_=LineRoundDotSetting):
            return self['round_size']

    @property
    def space_size(self) -> ap.Int:
        """
        Get a space size setting.

        Returns
        -------
        space_size : Int
            Blank space size between dots.
        """
        with ap.DebugInfo(
                callable_='space_size', locals_=locals(),
                module_name=__name__, class_=LineRoundDotSetting):
            return self['space_size']



"""Class implementation for graphic base class.
"""

from typing import Union

import apysc as ap
from apysc._display.display_object import DisplayObject
from apysc._display.flip_x_interface import FlipXInterface
from apysc._display.flip_y_interface import FlipYInterface
from apysc._display.rotation_around_center_interface import \
    RotationAroundCenterInterface
from apysc._display.rotation_around_point_interface import \
    RotationAroundPointInterface
from apysc._display.scale_x_from_center_interface import \
    ScaleXFromCenterInterface
from apysc._display.scale_x_from_point_interface import \
    ScaleXFromPointInterface
from apysc._display.scale_y_from_center_interface import \
    ScaleYFromCenterInterface
from apysc._display.scale_y_from_point_interface import \
    ScaleYFromPointInterface
from apysc._display.skew_x_interface import SkewXInterface
from apysc._display.skew_y_interface import SkewYInterface


class GraphicsBase(
        DisplayObject, RotationAroundCenterInterface,
        RotationAroundPointInterface, ScaleXFromCenterInterface,
        ScaleYFromCenterInterface, ScaleXFromPointInterface,
        ScaleYFromPointInterface, FlipXInterface, FlipYInterface,
        SkewXInterface, SkewYInterface):

    _variable_name: str

    def __init__(
            self, parent: 'ap.Graphics',
            x: Union[int, ap.Int],
            y: Union[int, ap.Int],
            variable_name: str) -> None:
        """
        Vector graphic base class.

        Parameters
        ----------
        parent : Graphics
            Parent `Graphics` instance.
        x : int or Int
            X position.
        y : int or Int
            Y position.
        variable_name : str
            Variable name of this instance. This will be used to
            js expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=GraphicsBase):
            from apysc._display.graphics import Graphics
            from apysc._validation import display_validation
            from apysc._validation import number_validation
            from apysc._validation import string_validation

            display_validation.validate_graphics(graphics=parent)
            self.parent_graphics: Graphics = parent
            number_validation.validate_integer(integer=x)
            number_validation.validate_integer(integer=y)
            if isinstance(x, int):
                x = ap.Int(x)
            self._x = x
            if isinstance(y, int):
                y = ap.Int(y)
            self._y = y
            string_validation.validate_not_empty_string(string=variable_name)
            super(GraphicsBase, self).__init__(
                stage=self.parent_graphics.stage,
                variable_name=variable_name)



"""Class Implementation for basic event.
"""

from typing import Generic
from typing import Optional
from typing import TypeVar

from apysc._type.variable_name_interface import VariableNameInterface

T = TypeVar('T', bound=VariableNameInterface)


class Event(Generic[T], VariableNameInterface):
    """
    Basic event class.
    """

    _this: T

    def __init__(
            self, this: T,
            type_name: Optional[str] = None) -> None:
        """
        Basic event class.

        Parameters
        ----------
        this : VariableNameInterface
            Instance that listening event (e.g., Sprite).
        type_name : str or None, default None
            Type name to set. Only specify when inherit this class.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Event):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            self._validate_type_name_and_self_type(type_name=type_name)
            self._this = this
            if type_name is None:
                type_name = var_names.EVENT
            self.variable_name = expression_variables_util.\
                get_next_variable_name(type_name=type_name)

    def _validate_type_name_and_self_type(
            self, type_name: Optional[str]) -> None:
        """
        Validate type_name argument is None when self instance
        is not Event subclass, and the opposite pattern is true
        as well.

        Parameters
        ----------
        type_name : str or None
            Type name to set.

        Raises
        ------
        ValueError
            - If type_name is not None and self instance is Event type.
            - If type_name is None and self instance is not Event type.
        """
        from apysc._type import type_util
        if type_name is not None:
            if type_util.is_same_class_instance(Event, self):
                raise ValueError(
                    'type_name argument can be set only when this instance '
                    'is subclass of Event.')
            return
        if not type_util.is_same_class_instance(Event, self):
            raise ValueError(
                'type_name argument can\'t be set when this instance '
                'is Event (this will be used by Event subclass).')

    @property
    def this(self) -> T:
        """
        Get an instance that listening this event.

        Returns
        -------
        this : VariableNameInterface
            Instance that listening this event.
        """
        return self._this



"""Class implementation for the animation_rotation_aroound_center
interface.
"""

from typing import Union

import apysc as ap
from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.animation_rotation_around_center import \
    AnimationRotationAroundCenter
from apysc._animation.easing import Easing


class AnimationRotationAroundCenterInterface(AnimationInterfaceBase):

    def animation_rotation_around_center(
            self,
            rotation_around_center: Union[int, ap.Int],
            duration: Union[int, ap.Int] = 3000,
            delay: Union[int, ap.Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationRotationAroundCenter:
        """
        Set the rotation around the center animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        rotation_around_center : int or Int
            The final rotation of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_rotation_around_center : AnimationRotationAroundCenter
            Created animation setting instance.

        References
        ----------
        - animation_rotation_around_center interface document
            - https://bit.ly/3FLb6lK
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_rotation_around_center: AnimationRotationAroundCenter = \
            AnimationRotationAroundCenter(
                target=self,
                rotation_around_center=rotation_around_center,
                duration=duration,
                delay=delay,
                easing=easing)
        return animation_rotation_around_center



"""Class implementation for the timer.
"""

from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union

import apysc as ap
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._event.handler import HandlerData
from apysc._time.fps import FPS
from apysc._type.number_value_interface import NumberValueInterface
from apysc._type.variable_name_interface import VariableNameInterface

_O1 = TypeVar('_O1')
_O2 = TypeVar('_O2')
_TimerEvent = Any
_Handler = Callable[[_TimerEvent, _O1], None]


class Timer(VariableNameInterface, CustomEventInterface):
    """
    Timer class to handle function calling at regular intervals.

    References
    ----------
    - Timer document
        - https://simon-ritchie.github.io/apysc/timer.html
    - TimerEvent class document
        - https://simon-ritchie.github.io/apysc/timer_event.html
    - Timer class delay setting document
        - https://simon-ritchie.github.io/apysc/timer_delay.html
    - FPS enum document
        - https://simon-ritchie.github.io/apysc/fps.html
    - Timer class repeat_count setting
        - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
    """

    _delay: ap.Number
    _repeat_count: ap.Int
    _current_count: ap.Int
    _handler_data: HandlerData
    _handler_name: str
    _running: ap.Boolean

    def __init__(
            self,
            handler: _Handler[_O1],
            delay: Union[int, float, NumberValueInterface, FPS],
            repeat_count: Union[int, ap.Int] = 0,
            options: Optional[_O1] = None) -> None:
        """
        Timer class to handle function calling at regular intervals.

        Parameters
        ----------
        handler : _Handler
            A handler would be called at regular intervals.
        delay : int or float or Int or Number or FPS
            A delay between each handler's calling in milisecond
            or FPS value.
            If `FPS` value is specified, then this value will be
            a milisecond calculated with that FPS value (e.g., if
            the `FPS_60` value is specified, then `delay` will be
            16.6666667).
        repeat_count : int or Int
            Max count of a handler's calling. If the handler's calling
            count has reached this value, then a timer will stop.
            If 0 is specified, then a timer will loop forever.
        options : dict or None, default None
            Optional arguments dictionary to pass the the handler.

        References
        ----------
        - Timer document
            - https://simon-ritchie.github.io/apysc/timer.html
        - TimerEvent class document
            - https://simon-ritchie.github.io/apysc/timer_event.html
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html
        - FPS enum document
            - https://simon-ritchie.github.io/apysc/fps.html
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.handler import get_handler_name
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            from apysc._validation import number_validation
            from apysc._validation.handler_options_validation import \
                validate_options_type
            with TemporaryNotHandlerScope():
                validate_options_type(options=options)
                self.variable_name = \
                    expression_variables_util.get_next_variable_name(
                        type_name=var_names.TIMER)
                self._handler_name = get_handler_name(
                    handler=handler, instance=self)
                handler = self._wrap_handler(handler=handler)
                self._handler: _Handler[_O1] = handler
                delay = self._convert_delay_to_number(delay=delay)
                number_validation.validate_num_is_gte_zero(num=delay)
                self._delay = delay
                if not isinstance(repeat_count, ap.Int):
                    repeat_count = ap.Int(repeat_count)
                number_validation.validate_num_is_gte_zero(num=repeat_count)
                self._repeat_count = repeat_count
                self._running = ap.Boolean(False)
                self._current_count = ap.Int(0)
                if options is None:
                    options = {}  # type: ignore
                self._handler_data = {
                    'handler': self._handler,
                    'options': options,
                }
                ap.append_js_expression(
                    expression=f'var {self.variable_name};')

    def _convert_delay_to_number(
            self,
            delay: Union[int, float, NumberValueInterface, FPS]) -> ap.Number:
        """
        Convert each type of delay value to a `Number` value.

        Parameters
        ----------
        delay : int or float or Int or Number or FPS
            A delay between each handler's calling in milisecond
            or FPS value.

        Returns
        -------
        delay : Number
            Converted delay value.
        """
        with ap.DebugInfo(
                callable_=self._convert_delay_to_number, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._time.fps import FPSDefinition
            if isinstance(delay, FPS):
                fps_definition: FPSDefinition = delay.value
                delay = fps_definition.milisecond_intervals
            if not isinstance(delay, ap.Number):
                delay = ap.Number(delay)
            return delay

    @property
    def delay(self) -> ap.Number:
        """
        Get a delay value.

        Returns
        -------
        delay : Number
            A delay value of each handler's calling in milisecond.

        References
        ----------
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html
        """
        with ap.DebugInfo(
                callable_='delay', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._delay)

    @property
    def repeat_count(self) -> ap.Int:
        """
        Get a max count value of a handler's calling.

        Returns
        -------
        repeat_count : Int
            Max count of a handler's calling. If 0 is set, then a
            timer will loop forever.

        References
        ----------
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
        """
        with ap.DebugInfo(
                callable_='repeat_count', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._repeat_count)

    @property
    def running(self) -> ap.Boolean:
        """
        Get a boolean value whether this timer is running or not.

        Returns
        -------
        running : Boolean.
            A boolean value whether this timer is running or not.
        """
        with ap.DebugInfo(
                callable_='running', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._running)

    @property
    def current_count(self) -> ap.Int:
        """
        Get a current timer count.

        Returns
        -------
        current_count : Int
            A current timer count.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='current_count', locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._type import value_util
            return value_util.get_copy(value=self._current_count)

    def start(self) -> None:
        """
        Start this timer.

        References
        ----------
        - Timer class start and stop interfaces
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html
        """
        with ap.DebugInfo(
                callable_=self.start, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event import handler_circular_calling_util
            from apysc._event.handler import append_handler_expression
            from apysc._expression import expression_data_util
            from apysc._type import value_util
            delay_val_str: str = value_util.get_value_str_for_expression(
                value=self._delay)
            is_handler_circular_calling: bool = handler_circular_calling_util.\
                is_handler_circular_calling(handler_name=self._handler_name)
            if is_handler_circular_calling:
                handler_name: str = handler_circular_calling_util.\
                    get_prev_handler_name(handler_name=self._handler_name)
                variable_name: str = handler_circular_calling_util.\
                    get_prev_variable_name(handler_name=self._handler_name)
            else:
                handler_name = self._handler_name
                variable_name = self.variable_name
            expression: str = (
                f'\nif (_.isUndefined({variable_name})) {{'
                f'\n  {variable_name} = setInterval('
                f'{handler_name}, {delay_val_str});'
                '\n}'
            )
            expression_data_util.append_js_expression(expression=expression)

            e: ap.TimerEvent = ap.TimerEvent(this=self)
            append_handler_expression(
                handler_data=self._handler_data,
                handler_name=handler_name,
                e=e)
            self._running.value = True

    def _wrap_handler(self, handler: _Handler[_O1]) -> _Handler[_O1]:
        """
        Wrap a handler to update a current count value when
        it is called.

        Parameters
        ----------
        handler : _Handler
            Target handler.

        Returns
        -------
        wrapped : _Handler
            Wrapped handler.
        """

        def wrapped(e: ap.TimerEvent, options: Any) -> None:
            """
            Wrapped handler.

            Parameters
            ----------
            e : TimerEvent
                Event instance.
            options : dict
                Optional arguments dictionary.
            """
            e.this._current_count += 1
            e.this._current_count._value = 0
            handler(e, options)
            self._append_count_branch_expression()

        return wrapped

    def _append_count_branch_expression(self) -> None:
        """
        Append the timer stopping expression by the counting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_count_branch_expression,
                locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.custom_event_type import CustomEventType
            from apysc._expression import expression_data_util
            from apysc._expression.indent_num import Indent
            from apysc._type import value_util
            current_count_val_str: str = value_util.\
                get_value_str_for_expression(value=self._current_count)
            repeat_count_val_str: str = value_util.\
                get_value_str_for_expression(value=self._repeat_count)
            expression: str = (
                f'if ({repeat_count_val_str} !== 0 '
                f'&& {current_count_val_str} === {repeat_count_val_str}) {{'
                f'\n{self._get_stop_expression(indent_num=1)}'
            )
            expression_data_util.append_js_expression(expression=expression)
            with Indent():
                self._running.value = False
                self.trigger_custom_event(
                    custom_event_type=CustomEventType.TIMER_COMPLETE)
                self.unbind_custom_event_all(
                    custom_event_type=CustomEventType.TIMER_COMPLETE)
                self._current_count.value = 0
            expression = '\n}'
            expression_data_util.append_js_expression(expression=expression)

    def stop(self) -> None:
        """
        Stop this timer.

        - Timer class start and stop interfaces
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.stop, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._expression import expression_data_util
            self._running.value = False
            expression: str = self._get_stop_expression(indent_num=0)
            expression_data_util.append_js_expression(expression=expression)

    def _get_stop_expression(self, indent_num: int) -> str:
        """
        Get a stop interface expression string.

        Parameters
        ----------
        indent_num : int
            Indentation number to append.

        Returns
        -------
        expression : str
            Stop interface expression string.
        """
        with ap.DebugInfo(
                callable_=self._get_stop_expression, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._string import indent_util
            expression: str = (
                f'if (!_.isUndefined({self.variable_name})) {{'
                f'\n  clearInterval({self.variable_name});'
                f'\n  {self.variable_name} = undefined;'
                '\n}'
            )
            expression = indent_util.append_spaces_to_expression(
                expression=expression, indent_num=indent_num)
            return expression

    def reset(self) -> None:
        """
        Reset the timer count and stop this timer.

        References
        ----------
        - Timer class reset interface document
            - https://simon-ritchie.github.io/apysc/timer_reset.html
        """
        with ap.DebugInfo(
                callable_=self.reset, locals_=locals(),
                module_name=__name__, class_=Timer):
            self.stop()
            self._current_count.value = 0

    def timer_complete(
            self, handler: _Handler[_O2],
            options: Optional[_O2] = None) -> str:
        """
        Add a timer complete event listener setting.

        Parameters
        ----------
        handler : _Handler
            A callable will be called when a timer is complete.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Timer class timer_complete interface document
            - https://simon-ritchie.github.io/apysc/timer_complete.html
        - About the handler options’ type document
            - https://bit.ly/39tnYxC
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.timer_complete, locals_=locals(),
                module_name=__name__, class_=Timer):
            from apysc._event.custom_event_type import CustomEventType
            from apysc._validation.handler_options_validation import \
                validate_options_type
            validate_options_type(options=options)
            e: ap.TimerEvent = ap.TimerEvent(this=self)
            name: str = self.bind_custom_event(
                custom_event_type=CustomEventType.TIMER_COMPLETE,
                handler=handler,
                e=e,
                options=options)
            return name
